@@ -207,6 +207,9 @@ unsigned short syntax_table[256];
 
 /* Comment chars.  */
 STRING eolcomm;
+STRING lquote, rquote;
+
+boolean visible_quotes;
 
 
 
@@ -692,6 +695,21 @@ unget_input (int ch)
     push_single(ch);
 }
 
+/*---------------------------------------------------------------.
+| The function unget_string () puts back a string on the input   |
+| stack, using the unget_input () subroutine.  This function is  |
+| global because it is called by collect_body ().                |
+`---------------------------------------------------------------*/
+
+void
+unget_string (char *text)
+{
+  char *cp;
+
+  for (cp=text + strlen (text) - 1; cp>=text; cp--)
+    unget_input (*cp);
+}
+
 /*------------------------------------------------------------------------.
 | skip_line () simply discards all immediately following characters, upto |
 | the first newline.  All whitespaces following this newline are also     |
@@ -811,11 +829,18 @@ input_init (void)
   eolcomm.string = xstrdup (DEF_EOLCOMM);
   eolcomm.length = strlen (eolcomm.string);
 
+  lquote.string  = xstrdup (DEF_LQUOTE);
+  lquote.length  = strlen (lquote.string);
+
+  rquote.string  = xstrdup (DEF_RQUOTE);
+  rquote.length  = strlen (rquote.string);
+  visible_quotes = FALSE;
+
   for (ch = 256; --ch > 0; )
     {
       if (isspace(ch))
         set_syntax_internal(SYNTAX_SPACE, ch);
-      else if (isalpha(ch) || ch == '_' || ch == '%')
+      else if (isalpha(ch) || ch == '_')
         set_syntax_internal(SYNTAX_ALPHA, ch);
       else if (isdigit(ch))
         set_syntax_internal(SYNTAX_NUM, ch);
@@ -849,6 +874,8 @@ input_deallocate (void)
 {
   xfree (array_current_line);
   xfree (eolcomm.string);
+  xfree (lquote.string);
+  xfree (rquote.string);
 
   obstack_free (&token_stack, 0);
   obstack_free (&input_stack, 0);
@@ -923,7 +950,7 @@ set_syntax (int code, const char *chars)
 `-------------------------------------------------------------------------*/
 
 token_type
-next_token (token_data *td, read_type expansion)
+next_token (token_data *td, read_type expansion, boolean in_string)
 {
   int ch;
   token_type type = TOKEN_NONE;
@@ -966,42 +993,64 @@ next_token (token_data *td, read_type expansion)
       (void) next_char ();
       if (IS_ESCAPE(ch))             /* ESCAPED WORD */
         {
-          obstack_1grow (&token_stack, ch);
-          if ((ch = peek_input ()) != CHAR_EOF)
+          if (lquote.length > 0 && MATCH (ch, lquote.string))
             {
-              if (ch == '/')
+              if (visible_quotes || expansion == READ_ATTR_VERB
+                  || expansion == READ_ATTR_ASIS || expansion == READ_BODY)
+                obstack_grow (&token_stack, lquote.string, lquote.length);
+              while ((ch = next_char ()) != CHAR_EOF)
                 {
-                  obstack_1grow (&token_stack, '/');
-                  (void) next_char ();
-                  ch = peek_input ();
-                }
-              if (IS_ALPHA(ch))
-                {
-                  ch = next_char ();
+                  if (rquote.length > 0 && MATCH (ch, rquote.string))
+                    break;
                   obstack_1grow (&token_stack, ch);
-                  while ((ch = next_char ()) != CHAR_EOF && IS_ALNUM(ch))
+                }
+              if (visible_quotes || expansion == READ_ATTR_VERB
+                  || expansion == READ_ATTR_ASIS || expansion == READ_BODY)
+                {
+                  obstack_grow (&token_stack, rquote.string, rquote.length);
+                  type = TOKEN_STRING;
+                }
+              else
+                type = TOKEN_QUOTED;
+            }
+          else
+            {
+              obstack_1grow (&token_stack, ch);
+              if ((ch = peek_input ()) != CHAR_EOF)
+                {
+                  if (ch == '/')
                     {
-                      if (eolcomm.length > 0 && MATCH (ch, eolcomm.string))
-                        {
-                          skip_line ();
-                          ch = CHAR_EOF;
-                          break;
-                        }
-                      obstack_1grow (&token_stack, ch);
+                      obstack_1grow (&token_stack, '/');
+                      (void) next_char ();
+                      ch = peek_input ();
                     }
-                  unget_input(ch);
+                  if (IS_ALPHA(ch))
+                    {
+                      ch = next_char ();
+                      obstack_1grow (&token_stack, ch);
+                      while ((ch = next_char ()) != CHAR_EOF && IS_ALNUM(ch))
+                        {
+                          obstack_1grow (&token_stack, ch);
+                        }
+                      if (ch == '*')
+                        {
+                          obstack_1grow (&token_stack, ch);
+                          ch = peek_input ();
+                        }
+                      else
+                        unget_input(ch);
 
-                  ch = peek_input ();
-                  if (IS_SPACE(ch) || IS_CLOSE(ch))
-                    type = TOKEN_WORD;
+                      if (IS_SPACE(ch) || IS_CLOSE(ch) || ch == '/')
+                        type = TOKEN_WORD;
+                      else
+                        type = TOKEN_STRING;
+                    }
                   else
                     type = TOKEN_STRING;
                 }
               else
-                type = TOKEN_STRING;
+                type = TOKEN_SIMPLE;        /* escape before eof */
             }
-          else
-            type = TOKEN_SIMPLE;        /* escape before eof */
         }
       else if (IS_CLOSE(ch))
         {
@@ -1033,18 +1082,23 @@ next_token (token_data *td, read_type expansion)
           obstack_1grow (&token_stack, ch);
           while ((ch = next_char ()) != CHAR_EOF && (IS_ALNUM(ch)))
             {
-              if (eolcomm.length > 0 && MATCH (ch, eolcomm.string))
-                {
-                  skip_line ();
-                  ch = CHAR_EOF;
-                  break;
-                }
               obstack_1grow (&token_stack, ch);
             }
-          unget_input(ch);
+          if (ch == '*')
+            {
+              obstack_1grow (&token_stack, ch);
+              ch = peek_input ();
+            }
+          else
+            unget_input(ch);
 
           type = TOKEN_STRING;
         }
+      else if (IS_RQUOTE(ch))
+        {
+          MP4HERROR ((EXIT_FAILURE, 0,
+             _("INTERNAL ERROR: CHAR_RQUOTE found.")));
+        }
       else if (IS_LQUOTE(ch))             /* QUOTED STRING */
         {
           quote_level = 1;
@@ -1089,6 +1143,7 @@ next_token (token_data *td, read_type expansion)
               break;
 
             case READ_ATTR_ASIS:
+            case READ_ATTR_QUOT:
               obstack_1grow (&token_stack, '"');
               type = TOKEN_QUOTE;
               break;
@@ -1109,7 +1164,7 @@ next_token (token_data *td, read_type expansion)
               break;
 
             case READ_ATTRIBUTE:
-            case READ_ATTR_VERB:
+            case READ_ATTR_QUOT:
               ch = next_char();
               if (ch == 'n')
                 obstack_1grow (&token_stack, '\n');
@@ -1117,10 +1172,29 @@ next_token (token_data *td, read_type expansion)
                 obstack_1grow (&token_stack, '\t');
               else if (ch == 'r')
                 obstack_1grow (&token_stack, '\r');
-              else if (ch == '"')
+              else if (ch == '\\')
+                obstack_1grow (&token_stack, ch);
+              else if (ch == '"' && in_string)
                 obstack_1grow (&token_stack, CHAR_QUOTE);
               else
-                obstack_1grow (&token_stack, ch);
+                {
+                  if (!(exp_flags & EXP_STD_BSLASH))
+                    obstack_1grow (&token_stack, '\\');
+                  obstack_1grow (&token_stack, ch);
+                }
+
+              type = TOKEN_STRING;
+              break;
+
+            case READ_ATTR_VERB:
+              ch = next_char();
+              if (ch == '"' && in_string)
+                obstack_1grow (&token_stack, CHAR_QUOTE);
+              else
+                {
+                  obstack_1grow (&token_stack, '\\');
+                  obstack_1grow (&token_stack, ch);
+                }
 
               type = TOKEN_STRING;
               break;
@@ -1244,7 +1318,7 @@ lex_debug (void)
   token_type t;
   token_data td;
 
-  while ((t = next_token (&td, READ_NORMAL)) != TOKEN_EOF)
+  while ((t = next_token (&td, READ_NORMAL, FALSE)) != TOKEN_EOF)
     print_token ("lex", t, &td);
 }
 #endif
