@@ -34,8 +34,8 @@
 #include "mp4h.h"
 
 static void expand_macro __P ((symbol *, read_type));
+static void expand_unknown_tag __P ((char *, read_type));
 static void expand_token __P ((struct obstack *, read_type, token_type, token_data *));
-static void collect_body __P ((symbol *, struct obstack *, struct obstack *));
 
 /* Current recursion level in expand_macro ().  */
 int expansion_level = 0;
@@ -56,7 +56,7 @@ expand_input (void)
   token_type t;
   token_data td;
 
-  while ((t = next_token (&td, READ_NORMAL)) != TOKEN_EOF)
+  while ((t = next_token (&td, READ_NORMAL, FALSE)) != TOKEN_EOF)
     expand_token ((struct obstack *) NULL, READ_NORMAL, t, &td);
 }
 
@@ -100,18 +100,24 @@ expand_token (struct obstack *obs, read_type expansion, token_type t,
 
       /* macro names must begin with a letter, an underscore or a %.
          If another character is found, this string is not a
-         macro, it could be ePerl delimiters.  */
+         macro.   */
 
-      if (! IS_ALPHA (*text))
+      if (IS_ALPHA (*text) || (*text == '/' && IS_ALPHA (*(text+1))))
         {
-          shipout_text (obs, TOKEN_DATA_TEXT (td), strlen (TOKEN_DATA_TEXT (td)));
-          break;
+          sym = lookup_symbol (text, SYMBOL_LOOKUP);
+          if (sym == NULL || SYMBOL_TYPE (sym) == TOKEN_VOID)
+            {
+              if (exp_flags & EXP_NO_HTMLTAG)
+                shipout_text (obs, TOKEN_DATA_TEXT (td),
+                        strlen (TOKEN_DATA_TEXT (td)));
+              else
+                expand_unknown_tag (text, expansion);
+            }
+          else
+            expand_macro (sym, expansion);
         }
-      sym = lookup_symbol (text, SYMBOL_LOOKUP);
-      if (sym == NULL || SYMBOL_TYPE (sym) == TOKEN_VOID)
-        shipout_text (obs, TOKEN_DATA_TEXT (td), strlen (TOKEN_DATA_TEXT (td)));
       else
-        expand_macro (sym, expansion);
+        shipout_text (obs, TOKEN_DATA_TEXT (td), strlen (TOKEN_DATA_TEXT (td)));
       break;
 
     default:
@@ -148,7 +154,7 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp,
   /* Skip leading white space.  */
   do
     {
-      t = next_token (&td, expansion);
+      t = next_token (&td, expansion, FALSE);
     }
   while (t == TOKEN_SPACE);
 
@@ -195,20 +201,18 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp,
           break;
 
         case TOKEN_QUOTE:
-          if (group_level == 0)
-            in_string = !in_string;
-          if (expansion_level > 1)
-            obstack_grow (obs, TOKEN_DATA_TEXT (&td),
-                    strlen (TOKEN_DATA_TEXT (&td)));
+          in_string = !in_string;
+          obstack_grow (obs, TOKEN_DATA_TEXT (&td),
+                  strlen (TOKEN_DATA_TEXT (&td)));
           break;
 
         case TOKEN_QUOTED:
         case TOKEN_STRING:
-          if (t == TOKEN_QUOTED)
+          if (expansion_level > 0 && t == TOKEN_QUOTED)
             obstack_1grow (obs, CHAR_LQUOTE);
           obstack_grow (obs, TOKEN_DATA_TEXT (&td),
                   strlen (TOKEN_DATA_TEXT (&td)));
-          if (t == TOKEN_QUOTED)
+          if (expansion_level > 0 && t == TOKEN_QUOTED)
             obstack_1grow (obs, CHAR_RQUOTE);
           break;
 
@@ -231,8 +235,8 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp,
           abort ();
         }
 
-      *last_char_ptr = *(TOKEN_DATA_TEXT (&td) + strlen (TOKEN_DATA_TEXT (&td)) - 1);
-      t = next_token (&td, expansion);
+      *last_char_ptr = LAST_CHAR (TOKEN_DATA_TEXT (&td));
+      t = next_token (&td, expansion, in_string);
     }
 }
 
@@ -245,25 +249,25 @@ expand_argument (struct obstack *obs, read_type expansion, token_data *argp,
 `-------------------------------------------------------------------------*/
 
 static boolean
-collect_arguments (symbol *sym, read_type expansion, struct obstack *argptr,
-                   struct obstack *arguments)
+collect_arguments (char *symbol_name, read_type expansion,
+                   struct obstack *argptr, struct obstack *arguments)
 {
   int ch;
   token_data td;
   token_data *tdp;
   char *last_addr;
   boolean more_args;
-  char last_char;
+  char last_char = ' ';
 
   TOKEN_DATA_TYPE (&td) = TOKEN_TEXT;
-  TOKEN_DATA_TEXT (&td) = SYMBOL_NAME (sym);
+  TOKEN_DATA_TEXT (&td) = symbol_name;
   tdp = (token_data *) obstack_copy (arguments, (voidstar) &td, sizeof (td));
   obstack_grow (argptr, (voidstar) &tdp, sizeof (tdp));
 
   ch = peek_input ();
   if (IS_CLOSE (ch))
-    next_token (&td, expansion);
-  else if (IS_SPACE(ch) || IS_ESCAPE(ch))
+    (void) next_token (&td, READ_BODY, FALSE);
+  else if (IS_SPACE(ch) || IS_ESCAPE(ch) || ch == '/')
     {
       do
         {
@@ -275,7 +279,7 @@ collect_arguments (symbol *sym, read_type expansion, struct obstack *argptr,
             {
               MP4HERROR ((EXIT_FAILURE, 0,
                 _("ERROR:%s:%d: EOF when reading argument of the `%s' tag"),
-                     CURRENT_FILE_LINE, SYMBOL_NAME(sym)));
+                     CURRENT_FILE_LINE, symbol_name));
             }
           tdp = (token_data *)
             obstack_copy (arguments, (voidstar) &td, sizeof (td));
@@ -289,6 +293,12 @@ collect_arguments (symbol *sym, read_type expansion, struct obstack *argptr,
           && strlen (TOKEN_DATA_TEXT (tdp)) == 0)
         argptr->next_free = last_addr;
     }
+  else
+    {
+      MP4HERROR ((warning_status, 0,
+        _("INTERNAL ERROR: Bad tag expression in `%s'"),
+             CURRENT_FILE_LINE, symbol_name));
+    }
   return (last_char == '/');
 }
 
@@ -298,7 +308,8 @@ collect_arguments (symbol *sym, read_type expansion, struct obstack *argptr,
 | necessary to deal with nested expressions.                       |
 `-----------------------------------------------------------------*/
 static void
-collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
+collect_body (char *symbol_name, read_type expansion,
+              struct obstack *argptr, struct obstack *bodyptr)
 {
   token_type t;
   token_data td;
@@ -308,7 +319,7 @@ collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
 
   while (1)
     {
-      t = next_token (&td, READ_BODY);
+      t = next_token (&td, expansion, FALSE);
       text = TOKEN_DATA_TEXT (&td);
       switch (t)
         {                                /* TOKSW */
@@ -316,13 +327,17 @@ collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
         case TOKEN_MACDEF:
           MP4HERROR ((EXIT_FAILURE, 0,
             _("ERROR:%s:%d: EOF when reading body of the `%s' tag"),
-                 CURRENT_FILE_LINE, SYMBOL_NAME(sym)));
+                 CURRENT_FILE_LINE, symbol_name));
           break;
 
         case TOKEN_BGROUP:
         case TOKEN_EGROUP:
           break;
 
+        case TOKEN_QUOTE:
+          obstack_1grow (bodyptr, CHAR_QUOTE);
+          break;
+
         case TOKEN_QUOTED:
         case TOKEN_SIMPLE:
         case TOKEN_SPACE:
@@ -341,12 +356,12 @@ collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
           if (*text == '/')
             {
               text++;
-              if (strcasecmp (text, SYMBOL_NAME(sym)) == 0)
+              if (strcasecmp (text, symbol_name) == 0)
                 {
                   /*  gobble closing bracket */
                   do
                     {
-                      t = next_token (&td, READ_BODY);
+                      t = next_token (&td, expansion, FALSE);
                     }
                   while (! IS_CLOSE(*TOKEN_DATA_TEXT (&td)))
                     ;
@@ -364,12 +379,35 @@ collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
                 }
               else
                 {
+                  newsym = lookup_symbol (text, SYMBOL_LOOKUP);
+                  if (newsym)
+                    {
+                      if (!(exp_flags & EXP_NOWARN_NEST))
+                        MP4HERROR ((warning_status, 0,
+                          _("Warning:%s:%d: `</%s>' found when `</%s>' expected"),
+                               CURRENT_FILE_LINE, text, symbol_name));
+                      if (exp_flags & EXP_UNM_BREAK)
+                        {
+                          /*  Premature end of body parsing.  */
+                          obstack_1grow (bodyptr, '\0');
+                          TOKEN_DATA_TYPE (&td) = TOKEN_TEXT;
+                          TOKEN_DATA_TEXT (&td) = obstack_finish (bodyptr);
+                          tdp = (token_data *) obstack_copy (bodyptr,
+                              (voidstar) &td, sizeof (td));
+                          obstack_grow (argptr, (voidstar) &tdp, sizeof (tdp));
+
+                          /*  Push read input back on the stack  */
+                          unget_string (text-2);
+
+                          return;
+                        }
+                    }
                   obstack_grow (bodyptr, TOKEN_DATA_TEXT (&td),
                           strlen(TOKEN_DATA_TEXT (&td)) );
                   /*  gobble closing bracket */
                   do
                     {
-                      t = next_token (&td, READ_BODY);
+                      t = next_token (&td, expansion, FALSE);
                       obstack_grow (bodyptr, TOKEN_DATA_TEXT (&td),
                               strlen(TOKEN_DATA_TEXT (&td)) );
                     }
@@ -381,15 +419,21 @@ collect_body (symbol *sym, struct obstack *argptr, struct obstack *bodyptr)
             {
               newsym = lookup_symbol (text, SYMBOL_LOOKUP);
               if (newsym == NULL || SYMBOL_TYPE (newsym) == TOKEN_VOID)
-                shipout_text (bodyptr, TOKEN_DATA_TEXT (&td), strlen (TOKEN_DATA_TEXT (&td)));
+                {
+                  if (exp_flags & EXP_NO_HTMLTAG)
+                    shipout_text (bodyptr, TOKEN_DATA_TEXT (&td),
+                            strlen (TOKEN_DATA_TEXT (&td)));
+                  else
+                    expand_unknown_tag (text, expansion);
+                }
               else
-                expand_macro (newsym, READ_ATTR_ASIS);
+                expand_macro (newsym, expansion);
             }
           break;
 
         default:
           MP4HERROR ((warning_status, 0,
-            _("INTERNAL ERROR: Bad token type in collect_body ()")));
+            _("INTERNAL ERROR:%d: Bad token type in collect_body ()"), t));
           abort ();
         }
     }
@@ -474,20 +518,21 @@ expand_macro (symbol *sym, read_type expansion)
   if (traced && (debug_level & DEBUG_TRACE_CALL))
     trace_prepre (SYMBOL_NAME (sym), my_call_id);
 
-  if (expansion == READ_ATTR_ASIS || expansion == READ_BODY
-          || expansion == READ_ATTR_VERB)
+  if (expansion == READ_ATTR_ASIS || expansion == READ_ATTR_VERB
+      || expansion == READ_BODY)
     attr_expansion = READ_ATTR_ASIS;
   else if (! SYMBOL_EXPAND_ARGS (sym))
     attr_expansion = READ_ATTR_VERB;
   else
     attr_expansion = READ_ATTRIBUTE;
 
-  slash = collect_arguments (sym, attr_expansion, &argptr, &arguments);
+  slash = collect_arguments (SYMBOL_NAME (sym), attr_expansion, &argptr,
+                             &arguments);
   argc = obstack_object_size (&argptr) / sizeof (token_data *);
 
   if (SYMBOL_CONTAINER (sym))
     {
-      collect_body (sym, &argptr, &body);
+      collect_body (SYMBOL_NAME (sym), READ_BODY, &argptr, &body);
       argv = (token_data **) obstack_finish (&argptr);
     }
   else
@@ -498,16 +543,21 @@ expand_macro (symbol *sym, read_type expansion)
           cp = TOKEN_DATA_TEXT (argv[argc-1]);
           if (*cp == '/' && *(cp+1) == '\0')
             argc--;
-          else if (*(cp + strlen (cp) - 1) == '/')
-            *(cp + strlen (cp) - 1) = '\0';
+          else if (LAST_CHAR (cp) == '/')
+            LAST_CHAR (cp) = '\0';
         }
+      else if (!(exp_flags & EXP_NOWARN_SLASH))
+        MP4HERROR ((warning_status, 0,
+          _("Warning:%s:%d: `<%s>' requires a trailing slash"),
+               CURRENT_FILE_LINE, SYMBOL_NAME (sym)));
     }
 
   if (traced)
     trace_pre (SYMBOL_NAME (sym), my_call_id, argc, argv);
 
   obs_expansion = push_string_init ();
-  if (expansion == READ_NORMAL || expansion == READ_ATTRIBUTE)
+  if (expansion == READ_NORMAL || expansion == READ_ATTRIBUTE
+          || expansion == READ_ATTR_QUOT)
     {
       if (expansion_level > 1)
         obstack_1grow (obs_expansion, CHAR_BGROUP);
@@ -527,6 +577,14 @@ expand_macro (symbol *sym, read_type expansion)
             shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
           else
             {
+              /*  Attributes are put back on stack.
+                  The first idea is to write
+                    obstack_1grow (obs_expansion, CHAR_BGROUP);
+                    shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
+                    obstack_1grow (obs_expansion, CHAR_EGROUP);
+                  But this does not work because if string is `%attributes'
+                  then further expansion will be wrong.
+              */
               if (strlen (TOKEN_DATA_TEXT (argv[i])) == 0)
                 {
                   obstack_1grow (obs_expansion, CHAR_BGROUP);
@@ -542,6 +600,13 @@ expand_macro (symbol *sym, read_type expansion)
                 shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
             }
         }
+      if (!SYMBOL_CONTAINER (sym))
+        {
+          if (argc > 1)
+            obstack_1grow (obs_expansion, ' ');
+          obstack_1grow (obs_expansion, '/');
+        }
+
       obstack_1grow (obs_expansion, '>');
       if (SYMBOL_CONTAINER (sym))
         {
@@ -565,3 +630,106 @@ expand_macro (symbol *sym, read_type expansion)
   if (expansion_level == 0)
     clear_tag_attr ();
 }
+
+/*-------------------------------------------------------------------------.
+| This macro reads attributes without expanding macro.  It is useful to    |
+| print unknown tags.                                                      |
+`-------------------------------------------------------------------------*/
+
+static void
+expand_unknown_tag (char *name, read_type expansion)
+{
+  struct obstack arguments, argptr, body;
+  token_data **argv;
+  int argc, i;
+  struct obstack *obs_expansion;
+  const char *expanded;
+  char *symbol_name, *cp;
+  read_type attr_expansion;
+  boolean slash;
+
+  expansion_level++;
+
+  symbol_name = xstrdup (name);
+
+  obstack_init (&arguments);
+  obstack_init (&argptr);
+  obstack_init (&body);
+
+  if (expansion == READ_NORMAL || expansion == READ_ATTRIBUTE)
+    attr_expansion = READ_ATTR_QUOT;
+  else
+    attr_expansion = READ_ATTR_ASIS;
+
+  slash = collect_arguments (symbol_name, attr_expansion, &argptr, &arguments);
+  argc  = obstack_object_size (&argptr) / sizeof (token_data *);
+
+  /*
+     This tag is a simple tag if either
+         - tag name begins with a slash (i.e. this is an end tag)
+         - tag name last character is a star
+         - attributes are ended by a trailing slash
+  */
+  if (*(symbol_name) == '/')
+    slash = TRUE;
+
+  if (!slash && !(exp_flags & EXP_STAR_COMPLEX))
+    slash = (LAST_CHAR (symbol_name) == '*');
+
+  if (!slash && !(exp_flags & EXP_DFT_SIMPLE))
+    collect_body (symbol_name, expansion, &argptr, &body);
+
+  argv = (token_data **) obstack_finish (&argptr);
+
+  /*  When this tag is no more processed, remove the trailing slash.  */
+  if ((exp_flags & EXP_REMOVE_TRAILING_SLASH) &&
+      expansion == READ_NORMAL && slash)
+    {
+      cp = TOKEN_DATA_TEXT (argv[argc-1]);
+      if (*cp == '/' && *(cp+1) == '\0')
+        argc--;
+      else if (LAST_CHAR (cp) == '/')
+        LAST_CHAR (cp) = '\0';
+    }
+
+  if (expansion == READ_ATTR_ASIS
+     || expansion == READ_ATTR_VERB || expansion == READ_BODY)
+    expansion = READ_BODY;
+
+  obs_expansion = push_string_init ();
+
+  if (expansion != READ_BODY)
+    obstack_1grow (obs_expansion, CHAR_LQUOTE);
+  obstack_1grow (obs_expansion, '<');
+  shipout_string (obs_expansion, symbol_name, 0);
+
+  for (i = 1; i < argc; i++)
+    {
+      obstack_1grow (obs_expansion, ' ');
+      shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[i]), 0);
+    }
+  obstack_1grow (obs_expansion, '>');
+  if (expansion != READ_BODY)
+    obstack_1grow (obs_expansion, CHAR_RQUOTE);
+  if (!slash && !(exp_flags & EXP_DFT_SIMPLE))
+    {
+      shipout_string (obs_expansion, TOKEN_DATA_TEXT (argv[argc]), 0);
+      if (expansion != READ_BODY)
+        obstack_1grow (obs_expansion, CHAR_LQUOTE);
+      obstack_grow (obs_expansion, "</", 2);
+      shipout_string (obs_expansion, symbol_name, 0);
+      obstack_1grow (obs_expansion, '>');
+      if (expansion != READ_BODY)
+        obstack_1grow (obs_expansion, CHAR_RQUOTE);
+    }
+
+  expanded = push_string_finish (expansion);
+
+  --expansion_level;
+
+  obstack_free (&arguments, NULL);
+  obstack_free (&argptr, NULL);
+  obstack_free (&body, NULL);
+
+  xfree (symbol_name);
+}
