@@ -57,6 +57,11 @@ int warning_status = 0;
 /* Artificial limit for expansion_level in macro.c.  */
 int nesting_limit = 250;
 
+/* Flags to control how expansion is performed.  */
+#define DEFAULT_EXPANSION_FLAGS 3114
+
+int exp_flags = DEFAULT_EXPANSION_FLAGS;
+
 /* Name of frozen file to digest after initialization.  */
 const char *frozen_file_to_read = NULL;
 
@@ -104,6 +109,14 @@ xfree (voidstar p)
 | Print a usage message and exit with STATUS.  |
 `---------------------------------------------*/
 
+#define HELP_EXP_FLAGS(n,str) \
+    do                        \
+      {                       \
+        printf(" %c %4d %s\n", \
+                ((exp_flags & n) ? '*' : ' '), n, str);  \
+      }                       \
+    while (0)
+
 static void
 usage (int status)
 {
@@ -132,6 +145,23 @@ Preprocessor features:\n\
              stdout);
       fputs (_("\
 \n\
+Parser features:\n\
+  -X, --expansion=NUMBER       set parser behaviour according to the bits\n\
+                               of NUMBER, with (star marks current flags)\n\
+"), stdout);
+      HELP_EXP_FLAGS(    1, "do not parse unknown tags");
+      HELP_EXP_FLAGS(    2, "unknown tags are assumed being simple");
+      HELP_EXP_FLAGS(    4, "trailing star in tag name do not make this tag simple");
+      HELP_EXP_FLAGS(    8, "an unmatched end tag closes all previous unmatched begin tags");
+      HELP_EXP_FLAGS(   16, "interpret backslashes as printf");
+      HELP_EXP_FLAGS(   32, "remove trailing slash in tag attributes");
+      HELP_EXP_FLAGS(   64, "do not remove trailing star in tag name");
+      HELP_EXP_FLAGS(  128, "do not remove leading star in tag name");
+      HELP_EXP_FLAGS( 1024, "suppress warnings about bad nested tags");
+      HELP_EXP_FLAGS( 2048, "suppress warnings about missing trailing slash");
+
+      fputs (_("\
+\n\
 Limits control:\n\
   -H, --hashsize=PRIME         set symbol lookup hash table size\n\
   -L, --nesting-limit=NUMBER   change artificial nesting limit\n"),
@@ -182,6 +212,7 @@ static const struct option long_options[] =
   {"arglength", required_argument, NULL, 'l'},
   {"debug", optional_argument, NULL, 'd'},
   {"error-output", required_argument, NULL, 'o'},
+  {"expansion", required_argument, NULL, 'X'},
   {"fatal-warnings", no_argument, NULL, 'E'},
   {"freeze-state", required_argument, NULL, 'F'},
   {"hashsize", required_argument, NULL, 'H'},
@@ -203,7 +234,7 @@ static const struct option long_options[] =
   { 0, 0, 0, 0 },
 };
 
-#define OPTSTRING "D:EF:H:I:L:QR:U:d:hl:o:st:V"
+#define OPTSTRING "D:EF:H:I:L:QR:U:X:d:hl:o:st:V"
 
 int
 main (int argc, char *const *argv)
@@ -295,6 +326,15 @@ main (int argc, char *const *argv)
           }
         break;
 
+      case 'X':
+        exp_flags = atoi (optarg);
+        if (exp_flags < 0)
+          {
+            error (0, 0, _("Bad expansion flags: `%s'"), optarg);
+            exp_flags = DEFAULT_EXPANSION_FLAGS;
+          }
+        break;
+
       case 'l':
         max_debug_argument_length = atoi (optarg);
         if (max_debug_argument_length <= 0)
