@@ -136,6 +136,9 @@ char *mktemp ();
 #define _(Text) (Text)
 #endif
 
+/*  Last character of a string.  */
+#define LAST_CHAR(Text) *(Text + strlen (Text) - 1)
+
 
 /* Various declarations.  */
 
@@ -329,11 +332,36 @@ struct token_data
 /* The status of processing. */
 #define READ_NORMAL    (1 << 0)  /* normal expansion of macros */
 #define READ_ATTRIBUTE (1 << 1)  /* when reading macro arguments */
-#define READ_ATTR_VERB (1 << 2)  /* inside macros with attributes=verbatim */
-#define READ_ATTR_ASIS (1 << 3)  /* attributes are read without any
+#define READ_ATTR_QUOT (1 << 2)  /* like READ_ATTRIBUTE, but quotes
+                                    are preserved */
+#define READ_ATTR_VERB (1 << 3)  /* inside macros with attributes=verbatim */
+#define READ_ATTR_ASIS (1 << 4)  /* attributes are read without any
            modification, main difference with READ_ATTR_VERB is that quotes
-           are not removed and are part of this attribute  */
-#define READ_BODY      (1 << 4)  /* when reading body function */
+           and backslashes are not removed and are part of this attribute  */
+#define READ_BODY      (1 << 5)  /* when reading body function */
+
+/* Flags which determine how expansion is done  */
+#define EXP_NO_HTMLTAG   (1 << 0)  /* do not parse unknown tags */
+#define EXP_DFT_SIMPLE   (1 << 1)  /* HTML tags are simple */
+#define EXP_STAR_COMPLEX (1 << 2)  /* HTML tags whose last char is an asterisk
+                                      are by default simple tags, they become
+                                      complex when this flag is set.  */
+#define EXP_UNM_BREAK    (1 << 3)  /* An unmatched end tag closes all previous
+                                      unmatched begin tags.  */
+#define EXP_STD_BSLASH   (1 << 4)  /* By default, only 'n', 'r', 't', '"' and
+                                      '\\' are escaped.  When this flag is
+				      set, backslashes are interpreted as in
+				      printf.  */
+#define EXP_REMOVE_TRAILING_SLASH \
+                         (1 << 5)  /* Remove trailing slash in simple tag attributes */
+#define EXP_LEAVE_TRAILING_STAR \
+                         (1 << 6)  /* Do not remove trailing slash in simple tag attributes */
+#define EXP_LEAVE_LEADING_STAR \
+                         (1 << 7)  /* Do not remove trailing slash in simple tag attributes */
+#define EXP_NOWARN_NEST  (1 << 10) /* Suppress warning about bad nested tags */
+#define EXP_NOWARN_SLASH (1 << 11) /* Suppress warning about missing trailing slash */
+
+extern int exp_flags;
 
 typedef enum token_type token_type;
 typedef enum token_data_type token_data_type;
@@ -342,7 +370,7 @@ typedef int read_type;
 void input_init __P ((void));
 void input_deallocate __P ((void));
 int peek_input __P ((void));
-token_type next_token __P ((token_data *, read_type));
+token_type next_token __P ((token_data *, read_type, boolean));
 void skip_line __P ((void));
 void skip_buffer __P ((void));
 void input_close __P ((void));
@@ -355,6 +383,7 @@ struct obstack *push_string_init __P ((void));
 const char *push_string_finish __P ((read_type));
 void push_wrapup __P ((const char *));
 boolean pop_wrapup __P ((void));
+void unget_string __P ((char *));
 
 /* read a file verbatim */
 void read_file_verbatim __P ((struct obstack *));
@@ -367,6 +396,8 @@ extern int *array_current_line;
 
 /* Begin and end quote */
 extern STRING lquote, rquote;
+
+/* Eof-of-line comment  */
 extern STRING eolcomm;
 
 #define CHAR_LQUOTE '\1'
@@ -375,8 +406,13 @@ extern STRING eolcomm;
 #define CHAR_EGROUP '\4'
 #define CHAR_QUOTE  '\5'
 
+/* Default eof-of-line comment  */
 #define DEF_EOLCOMM  ";;;"
 
+/* Default quotes  */
+#define DEF_LQUOTE  "<@["
+#define DEF_RQUOTE  "]@>"
+
 /* Syntax table definitions. */
 /* Please read the comment at the top of input.c for details */
 extern unsigned short syntax_table[256];
@@ -409,12 +445,12 @@ extern unsigned short syntax_table[256];
 #define IS_ALPHA(ch)  ((syntax_table[(int)(ch)]&SYNTAX_VALUE) == SYNTAX_ALPHA)
 #define IS_NUM(ch)    ((syntax_table[(int)(ch)]&SYNTAX_VALUE) == SYNTAX_NUM)
 #define IS_ALNUM(ch)  ((((syntax_table[(int)(ch)]) & SYNTAX_ALNUM) != 0) \
-                          || ch == ':' || ch == '*' || ch == '-')
+                          || ch == ':' || ch == '-')
 
-#define IS_LQUOTE(ch) (ch == CHAR_LQUOTE)
-#define IS_RQUOTE(ch) (ch == CHAR_RQUOTE)
 #define IS_BGROUP(ch) (ch == CHAR_BGROUP)
 #define IS_EGROUP(ch) (ch == CHAR_EGROUP)
+#define IS_LQUOTE(ch) (ch == CHAR_LQUOTE)
+#define IS_RQUOTE(ch) (ch == CHAR_RQUOTE)
 
 void set_syntax __P ((int, const char *));
 void set_syntax_internal __P ((int, int));
@@ -433,6 +469,7 @@ void make_diversion __P ((int));
 void insert_diversion __P ((int));
 void insert_file __P ((FILE *));
 void freeze_diversions __P ((FILE *));
+int remove_special_chars __P ((char *, boolean));
 
 
 /* File symtab.c  --- symbol table definitions.  */
@@ -450,7 +487,6 @@ struct symbol
 {
   struct symbol *next;
   boolean traced;
-  boolean shadowed;
   boolean container;
   boolean expand_args;
 
@@ -462,7 +498,6 @@ struct symbol
 
 #define SYMBOL_NEXT(S)          ((S)->next)
 #define SYMBOL_TRACED(S)        ((S)->traced)
-#define SYMBOL_SHADOWED(S)      ((S)->shadowed)
 #define SYMBOL_CONTAINER(S)     ((S)->container)
 #define SYMBOL_EXPAND_ARGS(S)   ((S)->expand_args)
 #define SYMBOL_NAME(S)          ((S)->name)
@@ -508,6 +543,8 @@ struct builtin
 
 typedef struct builtin builtin;
 
+extern boolean visible_quotes;
+
 void locale_init __P ((void));
 void initialize_builtin __P ((symbol *));
 void builtin_init __P ((void));
